import 'dart:async';
import 'dart:developer';

import 'package:ocg_agora/src/model.dart';
import 'package:agora_rtc_engine/agora_rtc_engine.dart';
import 'package:flutter/material.dart';
import 'package:flutter_bloc/flutter_bloc.dart';
import 'package:permission_handler/permission_handler.dart';

import '../ocg_agora.dart';

class AgoraController {
  static bool _isInitiated = false;
  static BuildContext _context;
  static Timer _timer;
  static AgoraSettings _agoraSettings;

  // DEBUF
  static Future<void> initAgora(String appId, BuildContext context) async {
    await AgoraRtcEngine.create(appId);

    await AgoraRtcEngine.enableVideo();

    await AgoraRtcEngine.enableWebSdkInteroperability(true);
    await AgoraRtcEngine.setDefaultAudioRouteToSpeaker(false);
    await AgoraRtcEngine.setParameters(
        '''{\"che.video.lowBitRateStreamParameter\":{\"width\":640,\"height\":480,\"frameRate\":24,\"bitRate\":800}}''');
    _context = context;
    _isInitiated = true;
  }

  static BuildContext get context => _context;

  static int getPredefinedAgoraId(String userId) {
    return userId.hashCode;
  }

  static joinCall(String channelId, int agoraId,
      {AgoraSettings agoraSettings}) async {
    log("Bloc Join Call");

    // Getting Permisipns
    _agoraSettings = agoraSettings;

    Map<Permission, PermissionStatus> status =
        await [Permission.microphone, Permission.camera].request();
    int connectionState = await AgoraRtcEngine.getConnectionState();
    log("Connection $connectionState");
    if (connectionState == AgoraConnectionState.Connected ||
        connectionState == AgoraConnectionState.Connecting ||
        connectionState == AgoraConnectionState.Reconnecting) {
      throw Exception("Already in call");
    }
    if (connectionState == AgoraConnectionState.Aborted) {
      throw Exception("Aborted");
    }
    bool hasPermission = true;
    status.values.forEach((permissionStatus) {
      hasPermission =
          hasPermission && (permissionStatus == PermissionStatus.granted);
    });
    if (!hasPermission) {
      throw Exception("Permissions not granted");
    }

    //
    if (!_isInitiated) {
      return;
    }
    await Future.delayed(Duration(milliseconds: 500));
    await _applyProfile();
    BlocProvider.of<AgoraBloc>(context)
        .add(AgoraJoinAction(channelId, agoraId));
    // Handlers
    _joinChannel(channelId, agoraId);
    _callHandlers(channelId);

    return channelId;
  }

  static Future<void> endCall() async {
    try {
      endTimer();
      if (_isInitiated) {
        await AgoraRtcEngine.leaveChannel();
      }
    } catch (e) {
      log("Call Leave Error ", error: e);
      // log("Error", error: e);
      // BlocProvider.of<AgoraBloc>(context)
      //     .add(AgoraLeftAction(state.channelId, e.toString()));
    }
  }

  static Future<void> _applySettings() async {
    if (_agoraSettings != null) {
      AgoraState state = BlocProvider.of<AgoraBloc>(context).state;
      if (!(state is AgoraCurrentState)) {
        return;
      }
      // Settings

      AgoraCurrentState currentState = state;
      int agoraId = currentState.currentId;
      await AgoraController.mute(
          agoraId, (_agoraSettings.isAudioMuted ?? false), true);
      await AgoraController.muteVideo(
          agoraId, (_agoraSettings.isVideoMuted ?? false), true);
      await turnSpeaker(_agoraSettings.isSpeakerOn ?? false);
    }
  }

  static Future<void> _applyProfile() async {
    if (_agoraSettings.isBroadcast ?? false) {
      await AgoraRtcEngine.setChannelProfile(ChannelProfile.LiveBroadcasting);
      if (_agoraSettings.isBroadcastHost) {
        await AgoraRtcEngine.setClientRole(ClientRole.Broadcaster);
      } else {
        await AgoraRtcEngine.setClientRole(ClientRole.Audience);
      }
    } else {
      await AgoraRtcEngine.setChannelProfile(ChannelProfile.Communication);
    }
  }

  static Future<void> mute(int userId, bool mute, bool isLocal) async {
    if (isLocal) {
      await AgoraRtcEngine.muteLocalAudioStream(mute);
      await AgoraRtcEngine.enableLocalAudio(!mute);
      AgoraState state = BlocProvider.of<AgoraBloc>(context).state;
      if (state is AgoraCurrentState) {
        BlocProvider.of<AgoraBloc>(context).add(AgoraAudioMutedAction(
            state.channelId, state.currentId, mute, isLocal));
      }
    } else
      await AgoraRtcEngine.muteRemoteAudioStream(userId, mute);
  }

  static Future<void> muteVideo(
      int userId, bool muteVideo, bool isLocal) async {
    if (isLocal) {
      await AgoraRtcEngine.muteLocalVideoStream(muteVideo);
      AgoraState state = BlocProvider.of<AgoraBloc>(context).state;
      if (state is AgoraCurrentState) {
        BlocProvider.of<AgoraBloc>(context).add(AgoraVideoMutedAction(
            state.channelId, state.currentId, muteVideo, isLocal));
      }
    } else {
      await AgoraRtcEngine.muteRemoteVideoStream(userId, muteVideo);
    }
  }

  static Future<void> turnSpeaker(bool isSpeakerOn) async {
    await AgoraRtcEngine.setEnableSpeakerphone(isSpeakerOn);
    if (context != null) {
      AgoraState state = BlocProvider.of<AgoraBloc>(context).state;
      BlocProvider.of<AgoraBloc>(context)
          .add(AgoraSpeakerMutedAction(state.channelId, isSpeakerOn));
    }
  }

  static Future<void> switchCamera() async {
    await AgoraRtcEngine.switchCamera();
  }

  static Future<bool> _joinChannel(
    String channelId,
    int userId,
  ) async {
    // DEBUG
    await AgoraRtcEngine.joinChannel(null, channelId, null, userId);
    return true;
  }

  static void _callHandlers(String callId) {
    AgoraRtcEngine.onError = (dynamic code) {
      log("Bloc Error $code");
      // AgoraState state = BlocProvider.of<AgoraBloc>(context).state;

      // BlocProvider.of<AgoraBloc>(context)
      //     .add(AgoraLeftAction(state.channelId, "$code"));
    };

    AgoraRtcEngine.onActiveSpeaker = (val) {
      AgoraState state = BlocProvider.of<AgoraBloc>(context).state;
      BlocProvider.of<AgoraBloc>(context)
          .add(AgoraSpeakerChangeAction(state.channelId, val));
    };

    AgoraRtcEngine.onLocalAudioStateChanged = (int error, int state) {};

    AgoraRtcEngine.onRejoinChannelSuccess = (channel, userId, val) {
      BlocProvider.of<AgoraBloc>(context)
          .add(AgoraJoinedAction(channel, userId));
    };

    AgoraRtcEngine.onJoinChannelSuccess = (String channel, userId, val) async {
      BlocProvider.of<AgoraBloc>(context)
          .add(AgoraJoinedAction(channel, userId));
      _applySettings();
    };

    AgoraRtcEngine.onLeaveChannel = () {
      AgoraState state = BlocProvider.of<AgoraBloc>(context).state;
      BlocProvider.of<AgoraBloc>(context)
          .add(AgoraLeftAction(state.channelId, null));
      log("Call Left Channel");
    };

    AgoraRtcEngine.onUserMuteAudio = (
      uid,
      val,
    ) {
      AgoraState state = BlocProvider.of<AgoraBloc>(context).state;
      BlocProvider.of<AgoraBloc>(context)
          .add(AgoraAudioMutedAction(state.channelId, uid, val, false));
    };

    AgoraRtcEngine.onUpdatedUserInfo = (info, val) {};

    AgoraRtcEngine.onRemoteAudioStateChanged =
        (uid, val, int reason, int elapsed) {
      AgoraState state = BlocProvider.of<AgoraBloc>(context).state;
      BlocProvider.of<AgoraBloc>(context).add(AgoraAudioMutedAction(
          state.channelId, uid, val != VideoState.Decoding, false));
    };
    AgoraRtcEngine.onRemoteVideoStateChanged =
        (uid, val, int reason, int elapsed) {
      AgoraState state = BlocProvider.of<AgoraBloc>(context).state;
      BlocProvider.of<AgoraBloc>(context).add(AgoraVideoMutedAction(
          state.channelId, uid, val != VideoState.Decoding, false));
    };

    AgoraRtcEngine.onUserJoined = (int uid, int elapsed) async {
      AgoraState state = BlocProvider.of<AgoraBloc>(context).state;
      BlocProvider.of<AgoraBloc>(context)
          .add(AgoraFriendJoinedAction(state.channelId, uid));
    };

    AgoraRtcEngine.onUserOffline = (int uid, int reason) {
      AgoraState state = BlocProvider.of<AgoraBloc>(context).state;
      BlocProvider.of<AgoraBloc>(context)
          .add(AgoraFriendLeftAction(state.channelId, uid));
    };
  }

  ///
  /// Start [Timer]
  ///
  static void runTimer() {
    log("Timer Started");
    _timer = Timer.periodic(Duration(seconds: 1), (timer) {
      AgoraState agoraState = BlocProvider.of<AgoraBloc>(context).state;
      if (agoraState is AgoraCurrentState) {
        BlocProvider.of<AgoraBloc>(context).add(AgoraTimerEvent(
            agoraState.channelId, agoraState.timerInSeconds + 1));
      } else {
        timer.cancel();
        timer = null;
      }
    });
  }

  ///
  /// Start [Timer]
  ///
  static void endTimer() {
    log("Timer Ended");
    if (_timer != null) {
      _timer.cancel();
      _timer = null;
      AgoraState agoraState = BlocProvider.of<AgoraBloc>(context).state;

      BlocProvider.of<AgoraBloc>(context)
          .add(AgoraTimerEvent(agoraState.channelId, 0));
    }
  }
}
